#include "Dimension.h"
#include "Constants.h"
#include "Datei.h"
#include "Game.h"
#include "Globals.h"

using namespace Framework;

#define MAX_VIEW_DISTANCE CHUNK_SIZE * 8

Dimension::Dimension( int id )
    : dimensionId( id ),
    chunks( new Trie<Chunk>() ),
    entities( new RCArray<Entity>() )
{}

Dimension::~Dimension()
{
    entities->release();
    chunks->release();
}

void Dimension::updateVisibility()
{
    bool changed = true;
    while( changed )
    {
        changed = false;
        cs.lock();
        for( auto chunk : chunkList )
        {
            if( chunk )
            {
                chunk->getThis();
                cs.unlock();
                changed |= chunk->updateVisibility();
                chunk->release();
                cs.lock();
            }
        }
        cs.unlock();
    }
}

void Dimension::getAddrOf( Punkt cPos, char* addr ) const
{
    *(int*)addr = cPos.x;
    *((int*)addr + 1) = cPos.y;
}

void Dimension::getAddrOfWorld( Punkt wPos, char* addr ) const
{
    if( wPos.x < 0 )
        wPos.x -= CHUNK_SIZE;
    if( wPos.y < 0 ) // needed because otherwise would (-8, -8) have the same adress as (8, 8)
        wPos.y -= CHUNK_SIZE;
    wPos /= CHUNK_SIZE;
    getAddrOf( wPos, addr );
}

Chunk* Dimension::zChunk( Punkt wPos ) const
{
    char addr[ 8 ];
    getAddrOfWorld( wPos, addr );
    return chunks->z( addr, 8 );
}

Framework::Either<Block*, int> Dimension::zBlock( Vec3<int> location )
{
    Chunk* c = zChunk( currentGame->getChunkCenter( location.x, location.y ) );
    if( c )
    {
        int x = location.x % CHUNK_SIZE;
        int y = location.y % CHUNK_SIZE;
        if( x < 0 )
            x += CHUNK_SIZE;
        if( y < 0 )
            y += CHUNK_SIZE;
        return c->zBlockAt( Vec3<int>( x, y, location.z ) );
    }
    return 0;
}

void Dimension::addEntity( Entity* entity )
{
    entities->add( entity );
}

void Dimension::setChunk( Chunk* chunk, Punkt center )
{
    char addr[ 8 ];
    getAddrOfWorld( center, addr );
    Chunk* old = chunks->z( addr, 8 );
    cs.lock();
    if( old )
    {
        for( int i = 0; i < chunkList.getEintragAnzahl(); i++ )
        {
            if( chunkList.get( i ) == old )
            {
                chunkList.remove( i );
                break;
            }
        }
    }
    chunks->set( addr, 8, chunk );
    cs.unlock();
    if( chunk )
    {
        chunkList.add( chunk );
        chunk->setAdded();
    }
    getAddrOfWorld( center + Punkt( CHUNK_SIZE, 0 ), addr );
    Chunk* zChunk = chunks->z( addr, 8 );
    if( zChunk )
    {
        zChunk->setNeighbor( WEST, chunk );
        if( chunk )
            chunk->setNeighbor( EAST, zChunk );
    }
    getAddrOfWorld( center + Punkt( -CHUNK_SIZE, 0 ), addr );
    zChunk = chunks->z( addr, 8 );
    if( zChunk )
    {
        zChunk->setNeighbor( EAST, chunk );
        if( chunk )
            chunk->setNeighbor( WEST, zChunk );
    }
    getAddrOfWorld( center + Punkt( 0, CHUNK_SIZE ), addr );
    zChunk = chunks->z( addr, 8 );
    if( zChunk )
    {
        zChunk->setNeighbor( NORTH, chunk );
        if( chunk )
            chunk->setNeighbor( SOUTH, zChunk );
    }
    getAddrOfWorld( center + Punkt( 0, -CHUNK_SIZE ), addr );
    zChunk = chunks->z( addr, 8 );
    if( zChunk )
    {
        zChunk->setNeighbor( SOUTH, chunk );
        if( chunk )
            chunk->setNeighbor( NORTH, zChunk );
    }
}

int Dimension::getDimensionId() const
{
    return dimensionId;
}

bool Dimension::hasChunck( int x, int y ) const
{
    return zChunk( Punkt( x, y ) );
}

void Dimension::removeDistantChunks( Punkt wPos )
{
    Array<int> removed;
    int index = 0;
    for( Chunk* chunk : chunkList )
    {
        if( (chunk->getCenter() - wPos).getLength() > MAX_VIEW_DISTANCE )
            removed.add( index, 0 );
        index++;
    }
    for( int i : removed )
    {
        Chunk* chunk = chunkList.get( i );
        chunk->prepareRemove();
        setChunk( 0, chunk->getCenter() );
    }
}